@@ -26,7 +26,7 @@
  * SUCH DAMAGE.
  */
 /*
- * $Id: draw.c,v 1.169 2000/03/05 07:26:30 nishida Exp $
+ * $Id: draw.c,v 1.185 2001/03/17 09:48:14 nishida Exp $
  */
 
 #include "mgp.h"
@@ -75,7 +75,7 @@ static void shrink __P((char *, u_int));
 #endif
 
 static void draw_string __P((struct render_state *, char *));
-static void draw_fragment __P((struct render_state *, u_char *, u_int, char *, int));
+static char *draw_fragment __P((struct render_state *, u_char *, u_int, char *, int));
 static int iskinsokuchar __P((u_int));
 static struct render_object *obj_alloc __P((struct render_state *state));
 static void obj_free __P((struct render_state *, struct render_object *));
@@ -127,6 +127,12 @@ static void predraw __P((struct render_state *));
 static void set_background_pixmap __P((struct ctrl *));
 static void get_background_pixmap __P((struct ctrl *, struct render_state *));
 static void regist_background_pixmap __P((XImageInfo *, Image *));
+#ifdef MNG
+static void process_anim __P((struct render_state *, struct ctrl *));
+static void obj_draw_anim __P((struct render_state *, 
+	u_int, u_int, struct render_object *));
+#endif
+static int valign = VL_BOTTOM;
 
 #define CHECK_CACHE {if (caching){caching = -1; return;}}
 
@@ -505,6 +511,11 @@ process_direc(state, seenpause)
 		break;
 
 	case CTL_BACK:
+		if (state->line){
+			fprintf(stderr, 
+				"warning: %%back directive should be put in the first line of the page. ignored.\n");
+			break;
+		}
 		back_color[caching] = cp->ctl_value;
 		bg_ctl = cp;	/*update later*/
 		break;
@@ -540,11 +551,11 @@ process_direc(state, seenpause)
 
 #ifdef FREETYPE
 	case CTL_TFONT:
-		tfc_setfont(cp->ctc_value, 0);
+		tfc_setfont(cp->ctc2_value1, 0, cp->ctc2_value2);
 		break;
 
 	case CTL_TMFONT:
-		tfc_setfont(cp->ctc_value, 1);
+		tfc_setfont(cp->ctc_value, 1, NULL);
 		break;
 
 #endif /*FREETYPE*/
@@ -574,6 +585,7 @@ process_direc(state, seenpause)
 				cp->ctm_ximagesize, cp->ctm_yimagesize, 0,
 				cp->ctm_zoomflag, 0);
 		}
+		state->brankline = 0;
 	    }
 		break;
 
@@ -675,7 +687,7 @@ process_direc(state, seenpause)
 
 	case CTL_LINEEND:
 		/* blank lines */
-		if (state->maxascent + state->maxdescent < 3) {	/*XXX*/
+		if (state->brankline) {	/*XXX*/
 			state->maxascent = char_size[caching];
 			state->maxdescent = VERT_GAP(char_size[caching]);
 		}
@@ -758,6 +770,29 @@ process_direc(state, seenpause)
 		}
 		break;
 
+	case CTL_CHARSET:
+		if (get_regid(cp->ctc_value) < 0){
+			fprintf(stderr, "invalid charset \"%s\". ignored\n", 
+				cp->ctc_value);
+			break;
+		}
+		strcpy(mgp_charset, cp->ctc_value);
+		break;
+
+#ifdef MNG
+	case CTL_ANIM:
+		if (state->align == AL_LEFTFILL0) {
+			state->align = AL_LEFTFILL1;
+			state->leftfillpos = state->linewidth;
+		}
+		process_anim(state, cp);
+		break;
+#endif
+
+	case CTL_VALIGN:
+		valign = cp->cti_value;
+		break;
+
 	default:
 		fprintf(stderr,
 			"undefined directive %d at page %d line %d:\n\t",
@@ -805,6 +840,7 @@ draw_line_start(state)
 	state->maxascent = 0;
 	state->maxdescent = 0;
 	state->linewidth = 0;
+	state->brankline = 1;
 	while ((obj = state->obj))
 		obj_free(state, obj);
 }
@@ -861,10 +897,15 @@ draw_line_end(state)
 	}
 
 	state->ypos += state->maxascent;
-	if (VERT_GAP(state->maxascent) < state->maxdescent)
+
+	/* 
+  	 * we should ignore height of images to calculate line gap.
+	 * suggested by Toru Terao 
+	 */ 
+	if (VERT_GAP(char_size[caching]) < state->maxdescent)
 		state->ypos += state->maxdescent;
 	else
-		state->ypos += VERT_GAP(state->maxascent);
+		state->ypos += VERT_GAP(char_size[caching]);
 	state->ypos += 2;
 }
 
@@ -982,10 +1023,23 @@ draw_string(state, data)
 	};
 #define RTAB96_MAX	(sizeof(rtab96)/sizeof(rtab96[0]))
 	static char *rtab9494[] = {
-		"jisx0208.1983-*",	/* ESC $ @ or ESC $ ( @ */
+		"jisx0208.1978-*",	/* ESC $ @ or ESC $ ( @ */
 		"gb2312.1980-*",	/* ESC $ A or ESC $ ( A */
 		"jisx0208.1983-*",	/* ESC $ B or ESC $ ( B */
 		"ksc5601.1987-*",	/* ESC $ ( C */
+		NULL,			/* D */
+		NULL,			/* E */
+		NULL,			/* F */
+		NULL,			/* G */
+		NULL,			/* H */
+		NULL,			/* I */
+		NULL,			/* J */
+		NULL,			/* K */
+		NULL,			/* L */
+		NULL,			/* M */
+		NULL,			/* N */
+		"jisx0213.2000-1",	/* ESC $ ( O */
+		"jisx0213.2000-2",	/* ESC $ ( P */
 	};
 #define RTAB9494_MAX	(sizeof(rtab9494)/sizeof(rtab9494[0]))
 	int charset16 = 0;
@@ -1060,13 +1114,13 @@ draw_string(state, data)
 			}
 		}
 
-		draw_fragment(state, p, q - p, registry, charset16);
+		q = draw_fragment(state, p, q - p, registry, charset16);
 
 		p = q;
 	}
 }
 
-static void
+static char *
 draw_fragment(state, p, len, registry, charset16)
 	struct render_state *state;
 	u_char *p;
@@ -1092,6 +1146,8 @@ draw_fragment(state, p, len, registry, charset16)
 
 	while (len) {
 		code = charset16 ? p[0] * 256 + p[1] : p[0];
+		if (code != ' ') 
+			state->brankline = 0; /* This isn't brankline */
 
 #if 0
 		if (code == ' ') {
@@ -1133,7 +1189,8 @@ draw_fragment(state, p, len, registry, charset16)
 #endif
 #ifdef FREETYPE_CHARSET16
 			if (!(mgp_flag & FL_NOFREETYPE)
-			 && strncmp(registry, "jisx0208.1983-", 14) == 0) {
+			 && (strncmp(registry, "jisx0208.1983-", 14) == 0 ||
+			     strncmp(registry, "jisx0213.2000-", 14) == 0)) {
 				if (tfc_get(code, char_size[caching], 1, registry,
 						charset16)){
 					mode = MODE_FREETYPE;
@@ -1228,6 +1285,11 @@ draw_fragment(state, p, len, registry, charset16)
 		len -= (charset16 ? 2 : 1);
 
 		state->linewidth += HORIZ_STEP(char_size[caching], char_len);
+		/* ukai */
+		if (!charset16 && 
+			state->linewidth + HORIZ_STEP(char_size[caching], char_len)
+		    	> state->width)
+		    break;
 	}
 
 	if (state->width - state->leftfillpos / 2 < state->linewidth
@@ -1276,6 +1338,7 @@ draw_fragment(state, p, len, registry, charset16)
 			draw_line_itemsize(state, tail->ascent, tail->descent);
 		}
 	}
+	return p;
 }
 
 static int
@@ -1348,6 +1411,10 @@ obj_free(state, obj)
 		if (obj->data.icon.xpoint)
 			free(obj->data.icon.xpoint);
 		break;
+#ifdef MNG
+	case O_ANIM:
+		break;
+#endif /* MNG */
 	}
 	free(obj);
 }
@@ -1747,7 +1814,7 @@ obj_draw(state, target, xpos, ypos)
 	}\
 	drawarea[areaindex].x = x;\
 	drawarea[areaindex].y = y - obj->data.X->ascent;\
-	drawarea[areaindex].width = obj->data.X->charlen+1;\
+	drawarea[areaindex].width = obj->data.X->xmax+1;\
 	drawarea[areaindex].height = obj->data.X->height+1;\
 	areaindex ++;\
 }
@@ -1954,6 +2021,11 @@ obj_draw(state, target, xpos, ypos)
 		y += ypos;
 #endif
 		switch (obj->type) {
+#ifdef MNG
+		case O_ANIM:
+			obj_draw_anim(state, x, y, obj);
+			break;
+#endif MNG /* MNG */
 		case O_IMAGE:
 			obj_draw_image(target, x, y, obj);
 			break;
@@ -2370,12 +2442,12 @@ x_setfont(xfont, csize, registry, truescalable)
 }
 
 static u_int
-draw_onechar_x(state, code, x, y, size, registry, lastchar)
+draw_onechar_x(state, code, x, y, size, argregistry, lastchar)
 	struct render_state *state;
 	u_int code;
 	int x, y;
 	int size;
-	char *registry;
+	char *argregistry;
 	int lastchar;
 {
 	u_int charlen;
@@ -2384,9 +2456,15 @@ draw_onechar_x(state, code, x, y, size, registry, lastchar)
 	XCharStruct *cs;
 	char *metricsource;
 	char *seed;
+	char *registry;
 
+	if (code >= 0xa0 && ((!argregistry || !argregistry[0]) && mgp_charset))
+		registry = mgp_charset;
+	else 
+		registry = argregistry;
 	seed = x_findseed(state, registry);
 	xfontstruct = x_setfont(seed, char_size[caching], registry, NULL);
+
 	if (xfontstruct == NULL)
 		return 0;
 
@@ -2739,7 +2817,19 @@ image_load(state, filename, numcolor, ximagesize, yimagesize, backflag, zoomflag
 		goto end;
 	}
 
-	draw_line_itemsize(state, 0, height * yzoomrate / 100);
+	switch(valign){
+	case VL_TOP:
+		draw_line_itemsize(state, 0, height * yzoomrate / 100);
+		break;
+	case VL_BOTTOM:
+		draw_line_itemsize(state, height * yzoomrate / 100, 0);
+		break;
+	case VL_CENTER:
+		draw_line_itemsize(state, height * yzoomrate / 200, 
+			height * yzoomrate / 200);
+		break;
+	}
+
 	if (centerflag)
 		image_posx = char_size[caching] / 2 - (width * xzoomrate / 100) / 2;
 	else
@@ -3002,6 +3092,7 @@ process_icon(state, cp)
 	}
 
 	cp = NULL;
+	state->brankline = 0;
 }
 
 static void
@@ -3044,7 +3135,9 @@ draw_bar(state, cp)
 	XFillRectangle(display, state->target, gcsbar, st, state->ypos + width, len, swidth);
 
 	state->ypos += width + swidth + VERT_GAP(char_size[caching]) / 2;
-	state->maxascent = width + swidth;
+	if (state->maxascent < width + swidth)
+		state->maxascent = width + swidth;
+	state->brankline = 0;
 }
 
 static void
@@ -3370,7 +3463,9 @@ reparent_child_window(child_window, x, y)
 		target = parent;
 	}
 	XReparentWindow(display, child_window, window, x, y);
+#if 0
 	XDestroyWindow(display, target);
+#endif
 }
 
 void
@@ -3379,7 +3474,7 @@ draw_reinit(state)
 {
 	/* invalidate the background image cache */
 
-	bg_ctl_last = NULL;
+	bg_ctl_last = bg_ctl_cache = NULL;
 	x_registerseed(state, NULL, NULL);
 }
 
@@ -3566,7 +3661,7 @@ epstoimage(state, epsfile, x, y, width, height, xzoom, yzoom)
 	fprintf(fp, "(%s) run\n", epsfile);
 	fprintf(fp, "showpage\n");
 	fprintf(fp, "quit\n");
-	fsync(fd);
+	fflush(fp);
 	fclose(fp);
 	signal(SIGPIPE, sigpipe_handler);
 
@@ -3841,10 +3936,15 @@ cache_page(state, page)
 	struct render_state *state;
 	int page;
 {
+	struct ctrl *tmp_bg_ctl;
+	int tmp_bgindex;
+
 	/* we don't need caching */
 	if (cached_page == page || page > maxpage || page <= 0)
 		return;
 
+	if (!page_attribute[page].pg_linenum) return;
+
 	XFlush(display);
 	memset(state, 0, sizeof(struct render_state));
 	state->target = cachewin;  /*XXX*/
@@ -3852,6 +3952,8 @@ cache_page(state, page)
 	state->height = window_height;
 	state->page = page;
 	caching = 1;
+	tmp_bg_ctl = bg_ctl;
+	tmp_bgindex = bgindex;
 	if (verbose){
 		printf("now caching %d page ...\n", page);
 		fflush(stdout);
@@ -3862,6 +3964,8 @@ cache_page(state, page)
 	}
 	caching = 0;
 	cached_page = page;
+	bg_ctl = tmp_bg_ctl;
+	bgindex = tmp_bgindex;
 }
 
 static void
@@ -3884,7 +3988,13 @@ set_from_cache(state)
 	XSetForeground(display, gcfore, fore_color[0]);
 	XSetBackground(display, gcfore, back_color[0]);
 	bg_ctl = bg_ctl_last = bg_ctl_cache;
-	set_background_pixmap(bg_ctl);
+	if (bg_ctl){
+		for (i = 0; i < MAXBGPIXMAP; i ++){
+			if (bgpixmap[i].ctl && ctlcmp(bg_ctl, bgpixmap[i].ctl) == 0)
+				bgindex = i;
+		}
+		set_background_pixmap(bg_ctl);
+	}
 
 	switch(cache_effect){
 		case 1:
@@ -3907,6 +4017,7 @@ reset_background_pixmap()
 	int	i = 0;
 
 	bg_ctl_last = NULL;
+	bg_ctl_cache = NULL;
 
 	for (i = 0; i < MAXBGPIXMAP; i ++) {
 		if (bgpixmap[i].image){
@@ -3992,7 +4103,8 @@ predraw(state)
 	struct render_state *state;
 {
 	if (!caching && cached_page != state->page 
-			&& page_attribute[state->page].pg_text){
+			&& page_attribute[state->page].pg_text
+			&& page_attribute[state->page].pg_linenum){
 		cache_page(&cache_state, state->page);
 		set_from_cache(state);
 		pcache_process(state->page);
@@ -4029,7 +4141,10 @@ get_background_pixmap(ctl, state)
 			back_gradation(state, &ctl->ct_val.ctrl_grad);
 			break;
 		case CTL_BACK:
+			break;
 		default:
+			fprintf(stderr, "fatal error in get_background_pixmap()\n");
+			cleanup(-1);
 			break;
 		}
 	}
@@ -4061,6 +4176,7 @@ regist_background_pixmap(ximageinfo, image)
 			bgpixmap[j +1].image = bgpixmap[j].image;
 			bgpixmap[j +1].ximageinfo = bgpixmap[j].ximageinfo;
 		}
+		bg_ctl_last = NULL;
 		i = 0;
 	}
 
@@ -4087,7 +4203,7 @@ set_background_pixmap(ctl)
 				break;	
 		}
 		if (i == MAXBGPIXMAP){
-			fprintf(stderr, "fatal error in set_background_image()\n");
+			fprintf(stderr, "fatal error in set_background_pixmap()\n");
 			cleanup(-1);
 		}
 		XSetWindowBackgroundPixmap(display, window, bgpixmap[i].pixmap);
@@ -4096,7 +4212,8 @@ set_background_pixmap(ctl)
 		XSetWindowBackground(display, window, ctl->ctl_value);
 		break;
 	default:
-		fprintf(stderr, "fatal error in set_background_image()\n");
+		fprintf(stderr, "fatal error in set_background_pixmap() op=%d\n", 
+			ctl->ct_op);
 		cleanup(-1);
 		break;
 	}
@@ -4122,7 +4239,7 @@ XClearPixmap(display, target)
 				break;	
 		}
 		if (i == MAXBGPIXMAP){
-			fprintf(stderr, "fatal error in set_background_image()\n");
+			fprintf(stderr, "fatal error in XClearPixmap()\n");
 			cleanup(-1);
 		}
 
@@ -4139,8 +4256,102 @@ XClearPixmap(display, target)
 			gc_cache, 0, 0, window_width, window_height);
 		break;
 	default:
-		fprintf(stderr, "fatal error in set_background_image()\n");
+		fprintf(stderr, "fatal error in XClearPixmap()\n");
 		cleanup(-1);
 		break;
 	}
 }
+
+int
+get_regid(registry)
+    char *registry;
+{
+	char *p;
+
+	if (!registry || registry[0] == '\0') return 0;
+	if (strlen(registry) == 9 && !strncmp("iso8859-", registry, 8) &&
+		registry[8] >= '1' && registry[8] <= '4') {
+			p = registry + 8;
+			return atoi(p) -1;
+	} else
+		return -1;
+}
+
+#ifdef MNG
+int
+obj_new_anim(state, x, y, width, height, filename, key)
+    struct render_state *state;
+    int x, y;
+    int width, height;
+    char *filename;
+	void *key;
+{
+	struct render_object *obj;
+
+	obj = obj_alloc(state);
+	if (obj == NULL)
+		return 0;
+	obj->x = x;
+	obj->y = y;
+	obj->type = O_ANIM;
+	obj->data.anim.width = width;
+	obj->data.anim.height = height;
+	obj->data.anim.filename = strdup(filename);
+	obj->data.anim.key = key;	/* for regchild */
+	obj->ascent = 0;	/*XXX*/
+	obj->descent = height;	/*XXX*/
+	obj->vertloc = VL_CENTER;
+
+	return 1;
+}
+
+static void
+obj_draw_anim(state, x, y, obj)
+    struct render_state *state;
+	u_int x, y;
+	struct render_object *obj;
+{
+	pid_t 	pid;
+
+	if (!(pid = fork())){
+		mngload(obj->data.anim.filename, x, y, 
+			obj->data.anim.width, obj->data.anim.height);
+		while(1) sleep(1);
+	}
+	regchild(pid, obj->data.anim.key, -1, state->page);
+}
+
+static void
+process_anim(state, cp)
+    struct render_state *state;
+	struct ctrl *cp;
+{
+	int width, height;
+
+	width = 200;
+	height = 100;
+
+	/*
+	 * we support only mng so far
+	 */
+	mngpreload(state, cp->ctc_value, &width, &height);
+
+	obj_new_anim(state, state->linewidth, - height, 
+		width, height, cp->ctc_value, cp);
+
+	switch(valign){
+	case VL_TOP:
+		draw_line_itemsize(state, 0, height);
+		break;
+	case VL_BOTTOM:
+		draw_line_itemsize(state, height, 0);
+		break;
+	case VL_CENTER:
+		draw_line_itemsize(state, height /2 , height /2);
+		break;
+	}
+
+	state->linewidth += width;
+	state->brankline = 0;
+}
+#endif
