@@ -26,7 +26,7 @@
  * SUCH DAMAGE.
  */
 /*
- * $Id: mgp.h,v 1.118 2000/03/05 07:26:36 nishida Exp $
+ * $Id: mgp.h,v 1.126 2001/02/03 18:46:50 nishida Exp $
  */
 
 #include <stdio.h>
@@ -386,6 +386,7 @@ struct render_state {
 	int maxascent;
 	int maxdescent;
 	u_int linewidth;
+	u_int brankline;
 	struct ctrl *xfont;
 	struct render_object *obj, *objlast;
 };
@@ -410,6 +411,9 @@ struct render_object {
 #endif /* FREETYPE */
 		O_XFONT,
 		O_IMAGE,
+#ifdef MNG
+		O_ANIM,
+#endif /* MNG */
 		O_ICON
 	} type;
 	u_long fore;
@@ -436,6 +440,14 @@ struct render_object {
 			u_int npoint;
 			XPoint *xpoint;
 		} icon;
+#ifdef MNG
+		struct {
+			u_int width;
+			u_int height;
+			char *filename;
+			void *key;
+		} anim;
+#endif /* MNG */
 	} data;
 };
 
@@ -451,6 +463,7 @@ struct vfont {
 	u_int height;
 	u_int code;
 	u_int charlen;
+	u_int xmax;
 	int ascent;
 	int descent;
 	u_int xoff;
@@ -472,12 +485,14 @@ struct tfont {
 	u_int height;		/* bitmap height */
 	u_int code;
 	u_int charlen;		/* origin x axis advance width */
+	u_int xmax;			/* right edge of bbox */
 	int ascent;		/* (top of bitmap) - (origin y axis) */
 	int descent;		/* (origin y axis) - (bottom of bitmap) */
 	int xoff;		/* (left of bitmap) - (origin x axis) */
 	char *fontname;
 	u_char *dbitmap;
 	u_int ref;
+	int regid;		/* registry ID */
 };
 #endif /* FREETYPE */
 
@@ -540,6 +555,8 @@ extern u_long ctrl_color[2];
 extern u_int b_quality[2];
 extern u_int quality_flag;
 
+extern char mgp_charset[256];
+
 #define VERT_GAP(s)		((s) * vert_gap[caching] / 100)
 #define HORIZ_GAP(s)		((s) * horiz_gap[caching] / 100)
 #define VERT_STEP(s)		((s) + VERT_GAP(s))
@@ -594,6 +611,7 @@ extern char *freetypefont0;	/* font name to be used as a last resort */
 extern char *freetypemfont0;	/* font name to be used as a last resort */
 extern int unicode_map[65536];
 #endif
+extern int latin_unicode_map[3][256];
 #ifdef TTY_KEYINPUT
 extern volatile int ttykey_enable;
 #endif
@@ -624,6 +642,7 @@ extern void draw_page __P((struct render_state *, struct ctrl *));
 extern Bool draw_one __P((struct render_state *, XEvent *));
 extern void timebar __P((struct render_state *));
 extern void draw_reinit __P((struct render_state *));
+extern int get_regid __P((char *));
 
 /*parse.c*/
 extern void load_file __P((char *));
@@ -681,7 +700,7 @@ extern void image_zoomratio __P((struct render_state *, float *, float *, int, i
 /* tfont.c */
 extern int tfc_setsize __P((u_int));
 extern struct tfont *tfc_get __P((u_int, u_int, int, char *, int));
-extern void tfc_setfont __P((char *, int));
+extern void tfc_setfont __P((char *, int, char *));
 extern XImage *tfc_image __P((struct tfont *, long, long, XImage *, int, int));
 
 /* embed.c */
@@ -711,3 +730,4 @@ extern void usleep __P((u_int));
 #define DITHERED_BGRAD
 #define COLOR_BUGFIX
 #endif
+
