@@ -26,7 +26,7 @@
  * SUCH DAMAGE.
  */
 /*
- * $Id: print.c,v 1.86 1999/11/28 14:35:09 nishida Exp $
+ * $Id: print.c,v 1.94 2001/03/05 13:16:14 nishida Exp $
  */
 /*
  * Paper size selection code is based on psutil.c by Angus J. C. Duggan
@@ -53,6 +53,7 @@ static int colorps = 0;
 static int painticon = 0;
 static u_int curlinenum;
 static u_int curpagenum;
+static u_int valign = VL_BOTTOM;
 
 #ifdef FREETYPE
 static u_int usetfont = 0;
@@ -64,6 +65,7 @@ static struct imagepool {
 	int xsiz;
 	int ysiz;
 	int eps;
+	int target_text;    /* by A. Ito */
 } imagepool[256];		/*enough?*/
 static int nimagepool = 0;
 static struct textpool {
@@ -578,6 +580,7 @@ print_pageinit(page)
 {
 	fprintf(fp, "%%%%Page: %d %d\n", page, page);
 	fprintf(fp, "/ypos YMARGIN -1 mul def\n");
+	fprintf(fp, "/xpos XMARGIN def\n");
 	fprintf(fp, "initcharsize\n");
 }
 
@@ -751,6 +754,21 @@ process_direc_print(cp)
 		back = cp->ctl_value;
 		break;
 
+	/* 
+     * MARK and AGAIN processing by A.Ito, 14 Jul. 2000 
+     */
+	case CTL_MARK:
+		fprintf(fp,"/markx xpos def /marky ypos def\n");
+		break;
+
+	case CTL_AGAIN:
+		fprintf(fp,"markx marky moveto /xpos markx def /ypos marky def\n");
+		break;
+
+	case CTL_VALIGN:
+		valign = cp->cti_value;
+		break;
+
 	default:
 		break;
 	}
@@ -857,7 +875,32 @@ line_end()
 	fprintf(fp, "/totlen exch def\n");
 	fprintf(fp, "/inmargin true def /fillzero XMARGIN def\n");
 	fprintf(fp, "%sdefxpos ", alignstr);
-	fprintf(fp, "xpos ypos ymax sub moveto\n");
+
+	/*
+	 * to determine text base line, we need to set imgsize first. pretty ugly..
+	 */
+    {
+	struct ctrl *cp1;
+		for (i = 0; i < nimagepool; i++) {
+			cp1 = imagepool[i].image;
+			if (cp1) 
+				fprintf(fp, "%d setimgsize\n", imagepool[i].ysiz);	/*XXX*/
+		}
+	}
+	fprintf(fp, "/yypos ypos charsize imgsize gt \n");
+
+	switch(valign){
+	case VL_TOP:
+		fprintf(fp, " { 0 } { 0 } ifelse sub def\n");
+		break;
+	case VL_BOTTOM:
+		fprintf(fp, " { 0 } { imgsize charsize sub } ifelse sub def\n");
+		break;
+	case VL_CENTER:
+		fprintf(fp, " { 0 } { imgsize charsize sub 2 div } ifelse sub def\n");
+		break;
+	}
+	fprintf(fp, "xpos yypos ymax sub moveto\n");
 
 	for (i = 0; i < ntextpool; i++) {
 		if (textpool[i].text) {
@@ -870,7 +913,10 @@ line_end()
 		} else
 			fprintf(fp, "%d ", textpool[i].xsiz);
 		fprintf(fp, "updatetotlen ");
-		fprintf(fp, "%snewlinecheck ", alignstr);
+		if (textpool[i].text) 
+			fprintf(fp, "%snewlinecheck ", alignstr);
+		else
+			fprintf(fp, "pop ");
 
 		if (colorps
 		 && (i == 0
@@ -894,6 +940,7 @@ line_end()
 				fprintf(fp, "%d 0 rmoveto\n", textpool[i].xsiz);
 			}
 		}
+		fprintf(fp, "/xpos%d currentpoint pop def\n", i);
 	}
 
 	ntextpool = 0;
@@ -905,14 +952,26 @@ line_end()
 		if (mgp_flag & FL_VERBOSE)
 			fprintf(fp, "%% emit the content of imagepool\n");
 		cp1 = imagepool[i].image;
-		if (!cp1)
+#if 0
+		if (!cp1) {
+		        fprintf(fp, "/xpos xpos%d def xpos ypos moveto\n",
+			        imagepool[i].target_text);
 			continue;
-		fprintf(fp, "xpos %d add ypos moveto\n",
-			imagepool[i].xoffset);
+		}
 		image_load_print(cp1->ctm_fname, cp1->ctm_numcolor,
 			cp1->ctm_ximagesize, cp1->ctm_yimagesize, 0,
 			cp1->ctm_zoomflag);
 		fprintf(fp, "%d setimgsize\n", imagepool[i].ysiz);	/*XXX*/
+#else
+		if (cp1) {
+			image_load_print(cp1->ctm_fname, cp1->ctm_numcolor,
+				cp1->ctm_ximagesize, cp1->ctm_yimagesize, 0,
+				cp1->ctm_zoomflag);
+		}
+        fprintf(fp, "/xpos xpos%d def xpos ypos moveto\n",
+	        imagepool[i].target_text);
+#endif
+
 	}
 	nimagepool = 0;
     }
@@ -1222,6 +1281,11 @@ print_fragment(data, len, kanjimode, pfx)
 		*q++ = '\0';
 		text_remember(buf, ASCII, char_size[0], textstartpos, pfx);
 	}
+	/* by A.Ito 14 Jul. 2000 */
+	/* placeholder */
+	imagepool[nimagepool].image = NULL;
+	imagepool[nimagepool].target_text = ntextpool-1;
+	nimagepool++;
 }
 
 static char *
@@ -1302,50 +1366,39 @@ image_remember(cp, pool)
 	char *epsname;
 	Image *myimage;
 	float xzoom, yzoom;
+	int x1, y1, x2, y2, width, height, swidth, sheight;
 	struct render_state state;
 
-	if ((epsname = checkeps(cp->ctm_fname))) {
-		int x1, y1, x2, y2, width, height, swidth, sheight, xim, yim;
-
-		if (ps_boundingbox(epsname, &x1, &y1, &x2, &y2) < 0)
-			goto noneps;
+	if ((epsname = checkeps(cp->ctm_fname))){
+		if (ps_boundingbox(epsname, &x1, &y1, &x2, &y2) < 0) goto noneps; 
 
 		width = x2 - x1;
 		height = y2 - y1;
-		xim = (!cp->ctm_ximagesize ? 100 : cp->ctm_ximagesize);
-		yim = (!cp->ctm_yimagesize ? 100 : cp->ctm_yimagesize);
-
-#if 0
-		switch (cp->ctm_zoomflag) {
-		case 0:	/* relative to screen size */
-			swidth = window_width * xim / 100;
-			sheight = window_height * yim / 100;
-			break;
-		case 1:	/* relative to the loaded image */
-			swidth = width * xim / 100;
-			sheight = height * yim / 100;
-			break;
-		case 2:	/* relative to screen pixels: don't know how to handle */
-			swidth = width;
-			sheight = height;
-			break;
+	} else {
+		myimage = loadImage(cp->ctm_fname, mgp_flag & FL_VERBOSE);
+		if (!myimage) {
+			fprintf(stderr, "failed to open %s\n", cp->ctm_fname);
+			exit(1);
 		}
-#else
-		xzoom = (float) xim;
-		yzoom = (float) yim;
+		width = myimage->width;
+		height = myimage->height;
+		freeImage(myimage);
+	}
+		xzoom = (!cp->ctm_ximagesize ? 100 : cp->ctm_ximagesize);
+		yzoom = (!cp->ctm_yimagesize ? 100 : cp->ctm_yimagesize);
+
 		state.width = window_width;
 		state.height = window_height;
 		image_zoomratio(&state, &xzoom, &yzoom, cp->ctm_zoomflag,
 			width, height);
 		swidth = (int) width * xzoom / 100;
 		sheight = (int) height * yzoom / 100;
-#endif
+
 		pool->xsiz = swidth;
 		pool->ysiz = sheight;
 		pool->xoffset = linewidth;
 		pool->image = cp;
 		return;
-	}
 
 noneps:
 	myimage = loadImage(cp->ctm_fname, mgp_flag & FL_VERBOSE);
@@ -1369,7 +1422,7 @@ image_load_print(filename, numcolor, ximagesize, yimagesize, backflag, zoomflag)
 	int backflag;
 	int zoomflag;
 {
-	Image *myimage;
+	Image *myimage, *image;
 	Pixmap mypixmap;
 	XImageInfo *ximageinfo;
 	XImage	*print_image;
@@ -1433,6 +1486,11 @@ image_load_print(filename, numcolor, ximagesize, yimagesize, backflag, zoomflag)
 	state.height = window_height;
 	image_zoomratio(&state, &xzoomrate, &yzoomrate, zoomflag,
 		width, height);
+	image = myimage;
+	myimage = zoom(image, xzoomrate, yzoomrate, verbose);
+	freeImage(image);
+	width = myimage->width;
+	height = myimage->height;
 #endif
 
 	if (! (ximageinfo= imageToXImage(display, screen, visual, depth,
@@ -1454,10 +1512,10 @@ image_load_print(filename, numcolor, ximagesize, yimagesize, backflag, zoomflag)
 		xpos = 0;
 		break;
 	case AL_CENTER:
-		xpos = (window_width - width * xzoomrate /100) / 2;
+		xpos = (window_width - width) / 2;
 		break;
 	case AL_RIGHT:
-		xpos = window_width - width * xzoomrate /100;
+		xpos = window_width - width;
 		break;
 	}
 
@@ -1469,8 +1527,10 @@ image_load_print(filename, numcolor, ximagesize, yimagesize, backflag, zoomflag)
 					AllPlanes, ZPixmap);
 
 	fprintf(fp, "gsave\n");
-	fprintf(fp, "%d ypos %d sub translate\n",
-		xpos + paper_xmargin, print_height);
+
+	/* fixed by A.Ito */
+	fprintf(fp, "xpos ypos %d sub translate\n",
+		print_height); 
 	fprintf(fp, "%d %d scale\n", print_width, print_height);
 	fprintf(fp, "%d %d 8\n", print_width, print_height);
 	fprintf(fp, "[%d 0 0 %d 0 %d]\n", print_width,
@@ -1785,7 +1845,11 @@ print_eps(epsfp, filename, numcolor, ximagesize, yimagesize, backflag, zoomflag)
 	}
 
 	fprintf(fp, "BeginEPSF\n");
+/*
 	fprintf(fp, "%d XMARGIN add ypos translate\n", xpos);
+*/
+	/* by A.Ito */
+	fprintf(fp, "xpos ypos translate\n");
 	fprintf(fp, "%f %f scale\n", xscale, yscale);
 	fprintf(fp, "%d %d translate\n", -1 * x1, -1 * y2);
 	fprintf(fp, "%%%%BeginDocument: %s\n", filename);
